#include "glwidget.h"
#include "matrix.h"
#include <QMouseEvent>
#include <iostream>
GLWidget::GLWidget(QWidget *parent)
    : QGLWidget(parent),xRot(0), yRot(0), verticesLoft(NULL), normalsLoft(NULL),
    verticesCaps(NULL), normalsCaps(NULL), numVerticesLoft(0), numVerticesCaps(0)
{
    // add path vertices
    loftShape.push_back(Vector( 0.5,  0.0,  0.0));
    loftShape.push_back(Vector( 0.25, 0.0, -0.5));
    loftShape.push_back(Vector(-0.25, 0.0, -0.5));
    loftShape.push_back(Vector(-0.5,  0.0,  0.0));
    loftShape.push_back(Vector(-0.25, 0.0,  0.5));
    loftShape.push_back(Vector( 0.25, 0.0,  0.5));

    loftPath.push_back(Vector( 0.0, 0.0, 0.0));
    loftPath.push_back(Vector( 0.0, 2.0, 0.0));
    loftPath.push_back(Vector(-2.0, 6.0, 0.0));
    loftPath.push_back(Vector(-2.0, 8.0, 0.0));
    loftPath.push_back(Vector( 2.0, 8.0, 0.0));
}

GLWidget::~GLWidget()
{
}

void
GLWidget::changeLoftShape(int value)
{
    loftShape.clear();
    if (value == 1)
    {
        loftShape.push_back(Vector( 0.5, 0.0,  0.0));
        loftShape.push_back(Vector( 0.0, 0.0, -0.5));
        loftShape.push_back(Vector(-0.5, 0.0,  0.0));
        loftShape.push_back(Vector( 0.0, 0.0,  0.5));


    } else {
        loftShape.push_back(Vector( 0.5,  0.0,  0.0));
        loftShape.push_back(Vector( 0.25, 0.0, -0.5));
        loftShape.push_back(Vector(-0.25, 0.0, -0.5));
        loftShape.push_back(Vector(-0.5,  0.0,  0.0));
        loftShape.push_back(Vector(-0.25, 0.0,  0.5));
        loftShape.push_back(Vector( 0.25, 0.0,  0.5));
    }
    calculateVertexPositions();
    updateGL();
}

/* calculate cross product of two vectors and determine whether the second one
 is right or left of the first one */
static
bool isCCWRotation(const Vector& last, const Vector& current)
{
    return (last.cross(current)).z() > 0;
}

//The loft algorithm: The shape that lies on the x-z-area
//needs to be rotated at each control point of the loft path.
//Since every point of the loft path lies on the x-y-area,
//the dot product can be used to determine the angle
//the shape needs to be rotated around the z-axis.
//First the angle between the world-up-vector (0,1,0)
//and the direction vector between two control points
//(a path segment) is calculated for each path segment.
//Since it was requested to rotate around half the angle
//between two path segments, the angle between two path
//segments is calculated and the shape is rotated around
//half that angle in the opposite direction.
//The cross product/vector product is used to determine
//in which direction to rotate (clockwise or counter-clockwise).
void
GLWidget::calculateVertexPositions()
{
    std::vector<Vector> transformedPoints;
    const Vector worldUp(0,1,0);

    Vector lastPathSegDir = worldUp;
    for (size_t i=0; i<loftPath.size(); i++)
    {
        Vector pathSegmentDirection;
        if (i == loftPath.size() - 1)  //last control point of the loft path
        {
            pathSegmentDirection = lastPathSegDir;
        }
        else
            pathSegmentDirection = loftPath[i+1] - loftPath[i];

        float angle = worldUp.angle(pathSegmentDirection); //in radians
        float angle2 = lastPathSegDir.angle(pathSegmentDirection);

        if (!isCCWRotation(worldUp, pathSegmentDirection))
            angle = -angle;
        if (!isCCWRotation(lastPathSegDir, pathSegmentDirection))
            angle2 = -angle2;

        Matrix mat;
        mat.translate(loftPath[i]);

        mat.rotate(angle, Matrix::Z);
        mat.rotate(-angle2/2.0, Matrix::Z);

        for (size_t i=0; i<loftShape.size(); i++)
        {
            Vector transformed = mat * loftShape[i];
            transformedPoints.push_back(transformed);
        }

        lastPathSegDir = pathSegmentDirection;
    }
    calculateLoftQuadArray(transformedPoints);
    calculateCapsTriangleArray(transformedPoints);
}

void
GLWidget::calculateLoftQuadArray(const std::vector<Vector>& transformedPoints)
{
    if (verticesLoft)
        delete[] verticesLoft;
    if (normalsLoft)
        delete[] normalsLoft;

    numVerticesLoft = (loftPath.size()-1) * loftShape.size() * 4;
    verticesLoft = new float[numVerticesLoft * 3];//3 float values (x,y,z) for each vertex
    normalsLoft = new float[numVerticesLoft * 3];

    int it = 0;
    int itNormal = 0;
    for (size_t i=0; i<loftPath.size()-1; i++)
    {
        for (size_t j=0; j<loftShape.size(); j++)
        {
            int index[4];
            size_t lastPointindex = loftShape.size()-1;
            index[0] = j + (i*loftShape.size());
            index[1] = (j == lastPointindex ? 0 : j+1) + (i*loftShape.size());
            index[2] = (j == lastPointindex ? 0 : j+1) + ((i+1)*loftShape.size());
            index[3] = j + ((i+1)*loftShape.size());

            Vector tangent   = transformedPoints[index[1]] - transformedPoints[index[0]];
            Vector bitangent = transformedPoints[index[3]] - transformedPoints[index[0]];
            Vector normal = tangent.cross(bitangent);
            normal.normalize();

            for (int k=0; k<4; k++)
            {
                verticesLoft[it++] = transformedPoints[index[k]].x();
                verticesLoft[it++] = transformedPoints[index[k]].y();
                verticesLoft[it++] = transformedPoints[index[k]].z();
                normalsLoft[itNormal++] = normal.x();
                normalsLoft[itNormal++] = normal.y();
                normalsLoft[itNormal++] = normal.z();
            }
        }
    }
}

void
GLWidget::calculateCapsTriangleArray(const std::vector<Vector>& transformedPoints)
{
    if (verticesCaps)
        delete[] verticesCaps;
    if (normalsCaps)
        delete[] normalsCaps;

    numVerticesCaps = loftShape.size() * 2 * 3; //2 caps, 3 Vertices for each Triangle, 1 triangle for each shape vertex
    verticesCaps = new float[numVerticesCaps * 3];//x,y,z for each vertex
    normalsCaps  = new float[numVerticesCaps * 3];

    int startCapIndex = 0; //Index of the first vertex of the shape at start position of the path
    int endCapIndex = transformedPoints.size() - loftShape.size(); //index of the first vertex of the shape at end position of the path

    Vector startCapCenter = transformedPoints[startCapIndex]
        + 0.5 * (transformedPoints[startCapIndex + loftShape.size()/2]-transformedPoints[startCapIndex]); //center point of the cap at start pos

    Vector endCapCenter  = transformedPoints[endCapIndex]
        + 0.5 * (transformedPoints[endCapIndex + loftShape.size()/2]-transformedPoints[endCapIndex]); //center point of the cap at end pos

    int it = 0;
    int itNormals = 0;
    //start cap
    for (size_t i=0; i<loftShape.size(); i++)
    {
        int currentIndex = startCapIndex + i;
        int nextIndex = startCapIndex + (i == loftShape.size()-1 ? 0 : i+1);

        verticesCaps[it++] = startCapCenter.x();
        verticesCaps[it++] = startCapCenter.y();
        verticesCaps[it++] = startCapCenter.z();

        verticesCaps[it++] = transformedPoints[currentIndex].x();
        verticesCaps[it++] = transformedPoints[currentIndex].y();
        verticesCaps[it++] = transformedPoints[currentIndex].z();

        verticesCaps[it++] = transformedPoints[nextIndex].x();
        verticesCaps[it++] = transformedPoints[nextIndex].y();
        verticesCaps[it++] = transformedPoints[nextIndex].z();

        Vector tangent   = transformedPoints[currentIndex] - startCapCenter;
        Vector bitangent = transformedPoints[nextIndex] - startCapCenter;
        Vector normal    = bitangent.cross(tangent).normalize();

        for (int k=0; k<3; k++)
        {
            normalsCaps[itNormals++] = normal.x();
            normalsCaps[itNormals++] = normal.y();
            normalsCaps[itNormals++] = normal.z();
        }
    }
    //end cap
    for (size_t i=0; i<loftShape.size(); i++)
    {
        int currentIndex = endCapIndex + i;
        int nextIndex = endCapIndex + (i == loftShape.size()-1 ? 0 : i+1);

        verticesCaps[it++] = endCapCenter.x();
        verticesCaps[it++] = endCapCenter.y();
        verticesCaps[it++] = endCapCenter.z();

        verticesCaps[it++] = transformedPoints[currentIndex].x();
        verticesCaps[it++] = transformedPoints[currentIndex].y();
        verticesCaps[it++] = transformedPoints[currentIndex].z();

        verticesCaps[it++] = transformedPoints[nextIndex].x();
        verticesCaps[it++] = transformedPoints[nextIndex].y();
        verticesCaps[it++] = transformedPoints[nextIndex].z();

        Vector tangent   = transformedPoints[currentIndex] - endCapCenter;
        Vector bitangent = transformedPoints[nextIndex] - endCapCenter;
        Vector normal    = tangent.cross(bitangent).normalize();

        for (int k=0; k<3; k++)
        {
            normalsCaps[itNormals++] = normal.x();
            normalsCaps[itNormals++] = normal.y();
            normalsCaps[itNormals++] = normal.z();
        }
    }
}

void
GLWidget::initializeGL()
{
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glEnable(GL_COLOR_MATERIAL);
    glClearColor(0.8, 0.8, 0.8, 1.0);
    float params[4] = {1, 1, 0, 1};
    glLightfv(GL_LIGHT0, GL_POSITION, params);
    glEnable(GL_DEPTH_TEST);

    calculateVertexPositions();
}

void
GLWidget::resizeGL(int w, int h)
{
    glViewport(0, 0, w, h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
//    gluPerspective(60, static_cast<float>(w) / h, 0.1, 100);
    float zoomValue = 6;
    if (w < h)
    {
        float aspect = w / static_cast<float>(h);
        glOrtho(-zoomValue*aspect, zoomValue*aspect, -zoomValue, zoomValue, 0.1, 100);
    } else
    {
        float aspect = h / static_cast<float>(w);
        glOrtho(-zoomValue, zoomValue, -zoomValue*aspect, zoomValue*aspect, 0.1, 100);
    }
    glMatrixMode(GL_MODELVIEW);
}

void
GLWidget::paintGL()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();
    glTranslatef(0, -5, -10);

    glTranslatef(0, 5, 0);
    glRotatef(xRot, 1, 0, 0);
    glRotatef(yRot, 0, 1, 0);
    glTranslatef(0, -5, 0);

    glColor3f(1, 1, 0);

    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY);

    glVertexPointer(3, GL_FLOAT, 0, verticesLoft);
    glNormalPointer(GL_FLOAT, 0, normalsLoft);
    glDrawArrays(GL_QUADS, 0, numVerticesLoft);

    glVertexPointer(3, GL_FLOAT, 0, verticesCaps);
    glNormalPointer(GL_FLOAT, 0, normalsCaps);
    glDrawArrays(GL_TRIANGLES, 0, numVerticesCaps);

    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);
}

void GLWidget::mouseMoveEvent(QMouseEvent *event)
{
    int dx = event->x() - last.x();
    int dy = event->y() - last.y();

    if (event->buttons() & Qt::LeftButton)
    {
        int valX = (xRot + dy)%360;
        int valY = (yRot + dx)%360;

        this->xRot = valX;
        this->yRot = valY;
        updateGL();
    }
    last = event->pos();
}

void GLWidget::mousePressEvent(QMouseEvent *event)
{
    last = event->pos();
}
