#include "matrix.h"
#include <iostream>
#include <cmath>
#include <cstring>

/** generate an identity matrix */
static void makeIdentity(float* mat)
{
    mat[0] = 1;    mat[4] = 0;    mat[8]  = 0;    mat[12] = 0;
    mat[1] = 0;    mat[5] = 1;    mat[9]  = 0;    mat[13] = 0;
    mat[2] = 0;    mat[6] = 0;    mat[10] = 1;    mat[14] = 0;
    mat[3] = 0;    mat[7] = 0;    mat[11] = 0;    mat[15] = 1;
}

Matrix::Matrix()
{
    setIdentity();
}

Matrix::Matrix(const Vector &right, const Vector &up, const Vector &forward)
{
    mat[0] = right.x();    mat[4] = right.y();    mat[8]  = right.z();    mat[12] = 0;
    mat[1] = up.x();       mat[5] = up.y();       mat[9]  = up.z();       mat[13] = 0;
    mat[2] = forward.x();  mat[6] = forward.y();  mat[10] = forward.z();  mat[14] = 0;
    mat[3] = 0;            mat[7] = 0;            mat[11] = 0;            mat[15] = 1;
}

Matrix::~Matrix()
{
}

void
Matrix::setIdentity()
{
    makeIdentity(mat);
}

/** append a rotation to the current transformation matrix */
void
Matrix::rotate(float angle, Axis axis)
{
    float tempMat[16];
    float s = std::sin(angle);
    float c = std::cos(angle);

    switch(axis)
    {
    case X:
        tempMat[0] = 1; tempMat[4] = 0; tempMat[8]  = 0;  tempMat[12] = 0;
        tempMat[1] = 0; tempMat[5] = c; tempMat[9]  = -s; tempMat[13] = 0;
        tempMat[2] = 0; tempMat[6] = s; tempMat[10] = c;  tempMat[14] = 0;
        tempMat[3] = 0; tempMat[7] = 0; tempMat[11] = 0;  tempMat[15] = 1;
        break;
    case Y:
        tempMat[0] = c;  tempMat[4] = 0; tempMat[8]  = s; tempMat[12] = 0;
        tempMat[1] = 0;  tempMat[5] = 1; tempMat[9]  = 0; tempMat[13] = 0;
        tempMat[2] = -s; tempMat[6] = 0; tempMat[10] = c; tempMat[14] = 0;
        tempMat[3] = 0;  tempMat[7] = 0; tempMat[11] = 0; tempMat[15] = 1;
        break;
    case Z:
        tempMat[0] = c; tempMat[4] = -s; tempMat[8] = 0; tempMat[12] = 0;
        tempMat[1] = s; tempMat[5] = c; tempMat[9]  = 0; tempMat[13] = 0;
        tempMat[2] = 0; tempMat[6] = 0; tempMat[10] = 1; tempMat[14] = 0;
        tempMat[3] = 0; tempMat[7] = 0; tempMat[11] = 0; tempMat[15] = 1;
        break;
    }
    matrixMult(tempMat);
}

/** append a translastion to the current transformation matrix */
void
Matrix::translate(const Vector& v)
{
    float tempMat[16];
    makeIdentity(tempMat);

    tempMat[12] = v.x();
    tempMat[13] = v.y();
    tempMat[14] = v.z();

    matrixMult(tempMat);
}

/** append a scaling to the current transformation matrix */
void
Matrix::scale(const Vector& v)
{
    float tempMat[16];
    makeIdentity(tempMat);

    tempMat[0] = v.x();
    tempMat[5] = v.y();
    tempMat[10] = v.z();

    matrixMult(tempMat);
}


void
Matrix::matrixMult(float* otherMat)
{
    float tempMat[16];
    tempMat[0] = mat[0] * otherMat[0] + mat[4] * otherMat[1] + mat[8]  * otherMat[2] + mat[12] * otherMat[3];
    tempMat[1] = mat[1] * otherMat[0] + mat[5] * otherMat[1] + mat[9]  * otherMat[2] + mat[13] * otherMat[3];
    tempMat[2] = mat[2] * otherMat[0] + mat[6] * otherMat[1] + mat[10] * otherMat[2] + mat[14] * otherMat[3];
    tempMat[3] = mat[3] * otherMat[0] + mat[7] * otherMat[1] + mat[11] * otherMat[2] + mat[15] * otherMat[3];

    tempMat[4] = mat[0] * otherMat[4] + mat[4] * otherMat[5] + mat[8]  * otherMat[6] + mat[12] * otherMat[7];
    tempMat[5] = mat[1] * otherMat[4] + mat[5] * otherMat[5] + mat[9]  * otherMat[6] + mat[13] * otherMat[7];
    tempMat[6] = mat[2] * otherMat[4] + mat[6] * otherMat[5] + mat[10] * otherMat[6] + mat[14] * otherMat[7];
    tempMat[7] = mat[3] * otherMat[4] + mat[7] * otherMat[5] + mat[11] * otherMat[6] + mat[15] * otherMat[7];

    tempMat[8]  = mat[0] * otherMat[8] + mat[4] * otherMat[9] + mat[8]  * otherMat[10] + mat[12] * otherMat[11];
    tempMat[9]  = mat[1] * otherMat[8] + mat[5] * otherMat[9] + mat[9]  * otherMat[10] + mat[13] * otherMat[11];
    tempMat[10] = mat[2] * otherMat[8] + mat[6] * otherMat[9] + mat[10] * otherMat[10] + mat[14] * otherMat[11];
    tempMat[11] = mat[3] * otherMat[8] + mat[7] * otherMat[9] + mat[11] * otherMat[10] + mat[15] * otherMat[11];

    tempMat[12] = mat[0] * otherMat[12] + mat[4] * otherMat[13] + mat[8]  * otherMat[14] + mat[12] * otherMat[15];
    tempMat[13] = mat[1] * otherMat[12] + mat[5] * otherMat[13] + mat[9]  * otherMat[14] + mat[13] * otherMat[15];
    tempMat[14] = mat[2] * otherMat[12] + mat[6] * otherMat[13] + mat[10] * otherMat[14] + mat[14] * otherMat[15];
    tempMat[15] = mat[3] * otherMat[12] + mat[7] * otherMat[13] + mat[11] * otherMat[14] + mat[15] * otherMat[15];

    std::memcpy(&mat, &tempMat, sizeof(float)*16);
}

void
Matrix::mult(Matrix& other)
{
    matrixMult(other.mat);
}

Vector
Matrix::operator* (const Vector& v) const
{
    return Vector(mat[0] * v.x() + mat[4] * v.y() + mat[8]  * v.z() + mat[12] * v.w(),
                  mat[1] * v.x() + mat[5] * v.y() + mat[9]  * v.z() + mat[13] * v.w(),
                  mat[2] * v.x() + mat[6] * v.y() + mat[10] * v.z() + mat[14] * v.w(),
                  mat[3] * v.x() + mat[7] * v.y() + mat[11] * v.z() + mat[15] * v.w());
}

Vector
Matrix::right() const
{
    return Vector(mat[0], mat[4], mat[8]);
}

Vector
Matrix::up() const
{
    return Vector(mat[1], mat[5], mat[9]);
}

Vector
Matrix::forward() const
{
    return Vector(mat[2], mat[6], mat[10]);
}

void
Matrix::debugPrint() const
{
    for (int i=0; i<4; i++)
    {
        for (int j=0; j<4; j++)
        {
            std::cout << mat[j*4+i] << ", ";
        }
        std::cout << std::endl;
    }
}
