#ifndef VECTOR_H
#define VECTOR_H

#include <GL/gl.h>
#include <cmath>
#include <iostream>

//namespace CG
//{

/**
 ** A basic Vector class. Code is based on the btVector3 class
 ** defined by the Bullet physics engine (http://bulletphysics.com/Bullet/BulletFull/)
 */
class Vector
{
public:
    Vector(GLfloat x, GLfloat y, GLfloat z, GLfloat w=1.0);
    Vector();
    ~Vector();

    GLfloat x() const;
    GLfloat y() const;
    GLfloat z() const;
    GLfloat w() const;

    Vector& operator+= (const Vector& v);
    Vector& operator-= (const Vector& v);
    Vector& operator*= (GLfloat s);
    Vector& operator/= (GLfloat s);

    bool operator==(const Vector& v) const;
    bool operator!=(const Vector& v) const;
    GLfloat length() const;
    GLfloat length2() const;
    GLfloat dot(const Vector &v) const;
    Vector& normalize();
    Vector normalized() const;
    GLfloat angle(const Vector& v) const;
    GLfloat distance(const Vector& v) const;
    GLfloat distance2(const Vector& v) const;
    Vector cross(const Vector& v) const;

    void print() const;
    Vector lerp(const Vector &other, GLfloat t) const;

    friend Vector operator+ (const Vector& v1, const Vector& v2);
    friend Vector operator- (const Vector& v1, const Vector& v2);
    friend Vector operator* (const Vector& v1, const Vector& v2);
    friend Vector operator/ (const Vector& v1, const Vector& v2);
    friend Vector operator* (const Vector& v, GLfloat s);
    friend Vector operator/ (const Vector& v, GLfloat s);
    friend Vector operator* (GLfloat, const Vector& v);
    friend Vector operator- (const Vector& v);
private:
    GLfloat values[4];
};
//}

inline
Vector::Vector(GLfloat x, GLfloat y, GLfloat z, GLfloat w)
{
    values[0] = x;
    values[1] = y;
    values[2] = z;
    values[3] = w;
}

inline
Vector::Vector()
{
    values[0] = 0.0;
    values[1] = 0.0;
    values[2] = 0.0;
    values[3] = 1.0;
}

inline
Vector::~Vector()
{
}

inline
GLfloat
Vector::x() const
{
    return values[0];
}

inline
GLfloat
Vector::y() const
{
    return values[1];
}

inline
GLfloat
Vector::z() const
{
    return values[2];
}

inline
GLfloat
Vector::w() const
{
    return values[3];
}

inline
Vector&
Vector::operator+= (const Vector& v)
{
    values[0] += v.values[0];
    values[1] += v.values[1];
    values[2] += v.values[2];
    return *this;
}

inline
Vector&
Vector::operator-= (const Vector& v)
{
    values[0] -= v.values[0];
    values[1] -= v.values[1];
    values[2] -= v.values[2];
    return *this;
}

inline
Vector&
Vector::operator*= (GLfloat s)
{
    values[0] *= s;
    values[1] *= s;
    values[2] *= s;
    return *this;
}

inline
Vector&
Vector::operator/= (GLfloat s)
{
    if (s == 0.0)
    {
        std::cerr << "operator /=: Cannot divide by 0" << std::endl;
        return *this;
    }
    return *this *= 1.0 / s;
}

inline
bool
Vector::operator== (const Vector& other) const
{
    return ((values[3]==other.values[3]) &&
            (values[2]==other.values[2]) &&
            (values[1]==other.values[1]) &&
            (values[0]==other.values[0]));
}

inline
bool
Vector::operator!=(const Vector& other) const
{
    return !(*this == other);
}

inline
GLfloat
Vector::dot(const Vector& v) const
{
    return values[0] * v.values[0] + values[1] * v.values[1] + values[2] * v.values[2];
}

inline
GLfloat
Vector::length2() const
{
    return dot(*this);
}

inline
GLfloat
Vector::length() const
{
    return std::sqrt(length2());
}

inline
Vector&
Vector::normalize()
{
    return *this /= length();
}

inline
Vector
Vector::normalized() const
{
    return *this / length();
}

inline
GLfloat
Vector::angle(const Vector& v) const
{
    GLfloat s = std::sqrt(length2() * v.length2());
    if (s == 0.0)
    {
        std::cerr << "Angle: Cannot divide by 0" << std::endl;
        return 0.0;
    }

    return std::acos(dot(v) / s);
}

inline
GLfloat
Vector::distance2(const Vector& v) const
{
    return (v - *this).length2();
}

inline
GLfloat
Vector::distance(const Vector& v) const
{
    return (v - *this).length();
}

inline
Vector
Vector::cross(const Vector& v) const
{
    return Vector(
        values[1] * v.values[2] - values[2] * v.values[1],
        values[2] * v.values[0] - values[0] * v.values[2],
        values[0] * v.values[1] - values[1] * v.values[0]);
}

inline
Vector
operator+ (const Vector& v1, const Vector& v2)
{
    return Vector(v1.values[0] + v2.values[0],
                  v1.values[1] + v2.values[1],
                  v1.values[2] + v2.values[2]);
}

inline
Vector
operator* (const Vector& v1, const Vector& v2)
{
    return Vector(v1.values[0] * v2.values[0],
                  v1.values[1] * v2.values[1],
                  v1.values[2] * v2.values[2]);
}

inline
Vector
operator- (const Vector& v1, const Vector& v2)
{
    return Vector(v1.values[0] - v2.values[0],
                  v1.values[1] - v2.values[1],
                  v1.values[2] - v2.values[2]);
}

inline
Vector
operator- (const Vector& v)
{
    return Vector(-v.values[0],
                  -v.values[1],
                  -v.values[2]);
}

inline
Vector
operator* (const Vector& v, GLfloat s)
{
    return Vector(v.values[0] * s,
                  v.values[1] * s,
                  v.values[2] * s);
}

inline
Vector
operator* (GLfloat s, const Vector& v)
{
    return v * s;
}

inline
Vector
operator/ (const Vector& v, GLfloat s)
{
    if (s == 0.0)
    {
        std::cerr << "operator/(v, s).Cannot divide by 0" << std::endl;
        return Vector();
    }
    return v * (1.0 / s);
}

inline
Vector
operator/ (const Vector& v1, const Vector& v2)
{
    return Vector(v1.values[0] / v2.values[0],
                  v1.values[1] / v2.values[1],
                  v1.values[2] / v2.values[2]);
}

inline
void
Vector::print() const
{
    std::cout << values[0] << ", " << values[1] << ", " << values[2] << std::endl;
}

inline
Vector
Vector::lerp(const Vector& other, GLfloat t) const
{
    return Vector(values[0] + (other.values[0] - values[0]) * t,
                  values[1] + (other.values[1] - values[1]) * t,
                  values[2] + (other.values[2] - values[2]) * t);
}

#endif // VECTOR_H
